import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import book, fsd, pm
from ._ewing import ewing as stroud_cn_3_5
from ._hammer_stroud import hammer_stroud_2n as stroud_cn_5_2
from ._helpers import CnScheme
from ._mustard_lyness_blatt import mustard_lyness_blatt as stroud_cn_5_5
from ._phillips import phillips as stroud_cn_7_1
from ._stroud_1957 import stroud_1957_2 as stroud_cn_2_1
from ._stroud_1957 import stroud_1957_3 as stroud_cn_3_1
from ._stroud_1966 import stroud_1966_a as stroud_cn_5_4
from ._stroud_1966 import stroud_1966_b as stroud_cn_5_6
from ._stroud_1966 import stroud_1966_c as stroud_cn_5_7
from ._stroud_1966 import stroud_1966_d as stroud_cn_5_8
from ._stroud_1968 import stroud_1968 as stroud_cn_5_3
from ._thacher import thacher as stroud_cn_2_2
from ._tyler import tyler as stroud_cn_3_3

_source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_cn_1_1(n):
    # centroid scheme
    weights = np.array([1])
    points = np.full((1, n), 0)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud Cn 1-1", n, weights, points, 1, _source)


def stroud_cn_1_2(n):
    # product trapezoidal scheme
    weights = np.full(2 ** n, frac(1, 2 ** n))
    points = pm(n * [1])
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud Cn 1-2", n, weights, points, 1, _source, 1.777e-15)


def stroud_cn_3_2(n):
    weights = np.full(2 * n, frac(1, 2 * n))
    r = sqrt(frac(n, 3))
    points = fsd(n, (r, 1))
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud Cn 3-2", n, weights, points, 3, _source, 5.863e-14)


def stroud_cn_3_4(n):
    weights = np.full(2 ** n, frac(1, 2 ** n))
    r = sqrt(3) / 3
    points = pm(n * [r])
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud Cn 3-4", n, weights, points, 3, _source, 3.376e-14)


def stroud_cn_3_6(n):
    lst = n * [[frac(1, 3), frac(4, 3), frac(1, 3)]]
    weights = np.product(np.array(np.meshgrid(*lst)).T.reshape(-1, n), axis=-1)
    weights /= 2 ** n
    lst = n * [[-1, 0, +1]]
    points = np.array(np.meshgrid(*lst)).T.reshape(-1, n)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud Cn 3-6", n, weights, points, 3, _source, 1.221e-14)


def stroud_cn_5_9(n):
    lst = n * [[frac(5, 9), frac(8, 9), frac(5, 9)]]
    weights = np.product(np.array(np.meshgrid(*lst)).T.reshape(-1, n), axis=-1)
    weights /= 2 ** n
    sqrt35 = sqrt(frac(3, 5))
    lst = n * [[-sqrt35, 0, sqrt35]]
    points = np.array(np.meshgrid(*lst)).T.reshape(-1, n)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud Cn 5-9", n, weights, points, 5, _source, 1.224e-14)


__all__ = [
    "stroud_cn_1_1",
    "stroud_cn_1_2",
    "stroud_cn_2_1",
    "stroud_cn_2_2",
    "stroud_cn_3_1",
    "stroud_cn_3_2",
    "stroud_cn_3_3",
    "stroud_cn_3_4",
    "stroud_cn_3_5",
    "stroud_cn_3_6",
    # TODO implement Cn 5-1
    # Cn 5-1 is not implemented because it's based on explicit values only given for
    # n=4,5,6.
    "stroud_cn_5_2",
    "stroud_cn_5_3",
    "stroud_cn_5_4",
    "stroud_cn_5_5",
    "stroud_cn_5_6",
    "stroud_cn_5_7",
    "stroud_cn_5_8",
    "stroud_cn_5_9",
    "stroud_cn_7_1",
]



import numpy as np
from sympy import Rational as frac
from sympy import S, sqrt

from ..helpers import article, expand_symmetries, pm, untangle, z
from ._helpers import CnScheme, _fs11

_source = article(
    authors=["A.H. Stroud"],
    title="Some Fifth Degree Integration Formulas for Symmetric Regions",
    journal="Mathematics of Computation",
    volume="20",
    number="93",
    month="jan",
    year="1966",
    pages="90-97",
    url="https://doi.org/10.1090/S0025-5718-1966-0191094-8",
)


def stroud_1966_a(n):
    r = sqrt(frac(5 * n + 4, 30))
    s = sqrt(frac(5 * n + 4, 15 * n - 12))
    d = {
        "a0": [[frac(40, (5 * n + 4) ** 2)], [r]],
        "a": [[frac(5 * n - 4, (5 * n + 4)) ** 2 / 2 ** n], [s]],
    }
    points, weights = expand_symmetries(d, n)
    return CnScheme("Stroud 1966a", n, weights, points, 5, _source, 2.432e-14)


def stroud_1966_b(n):
    s = 1 / sqrt(3)
    data = [(frac(4, 5 * n + 4), z(n))]
    for k in range(1, n + 1):
        r = sqrt(frac(5 * k + 4, 15))
        arr = np.full((2 ** (n - k + 1), n), S(0))
        arr[:, k - 1 :] = pm((n - k + 1) * [1])
        arr[:, k - 1] *= r
        arr[:, k:] *= s
        b = frac(5 * 2 ** (k - n + 1), (5 * k - 1) * (5 * k + 4))
        data.append((b, arr))

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud 1966b", n, weights, points, 5, _source, 3.393e-14)


def stroud_1966_c(n):
    r = sqrt((5 * n + 4 + 2 * (n - 1) * sqrt(5 * n + 4)) / (15 * n))
    s = sqrt((5 * n + 4 - 2 * sqrt(5 * n + 4)) / (15 * n))
    data = [(frac(4, 5 * n + 4), z(n)), (frac(5, (5 * n + 4) * 2 ** n), _fs11(n, r, s))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud 1966c", n, weights, points, 5, _source, 2.097e-13)


def stroud_1966_d(n):
    r = sqrt((5 * n - 2 * sqrt(5) + 2 * (n - 1) * sqrt(5 * n + 5)) / (15 * n))
    # This sqrt() is imaginary for negative for n=2.
    s = sqrt((5 * n - 2 * sqrt(5) - 2 * sqrt(5 * n + 5)) / (15 * n))
    t = sqrt((5 + 2 * sqrt(5)) / 15)
    w = frac(1, 2 ** n * (n + 1))
    data = [(w, _fs11(n, r, s)), (w, pm(n * [t]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud 1966d", n, weights, points, 5, _source, 1.955e-14)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["F.D. Witherden", "P.E. Vincent"],
    title="On the identification of symmetric quadrature rules for finite element methods",
    journal="Computers & Mathematics with Applications",
    volume="69",
    number="10",
    month="may",
    year="2015",
    pages="1232–1241",
    url="https://doi.org/10.1016/j.camwa.2015.03.017",
)

this_dir = pathlib.Path(__file__).resolve().parent


def witherden_vincent_01():
    return _read(this_dir / "wv01.json", source)


def witherden_vincent_02():
    return _read(this_dir / "wv02.json", source)


def witherden_vincent_04():
    return _read(this_dir / "wv04.json", source)


def witherden_vincent_05():
    return _read(this_dir / "wv05.json", source)


def witherden_vincent_06():
    return _read(this_dir / "wv06.json", source)


def witherden_vincent_07():
    return _read(this_dir / "wv07.json", source)


def witherden_vincent_08():
    return _read(this_dir / "wv08.json", source)


def witherden_vincent_09():
    return _read(this_dir / "wv09.json", source)


def witherden_vincent_10():
    return _read(this_dir / "wv10.json", source)


def witherden_vincent_11():
    return _read(this_dir / "wv11.json", source)


def witherden_vincent_12():
    return _read(this_dir / "wv12.json", source)


def witherden_vincent_13():
    return _read(this_dir / "wv13.json", source)


def witherden_vincent_14():
    return _read(this_dir / "wv14.json", source)


def witherden_vincent_15():
    return _read(this_dir / "wv15.json", source)


def witherden_vincent_16():
    return _read(this_dir / "wv16.json", source)


def witherden_vincent_17():
    return _read(this_dir / "wv17.json", source)


def witherden_vincent_18():
    return _read(this_dir / "wv18.json", source)


def witherden_vincent_19():
    return _read(this_dir / "wv19.json", source)


def witherden_vincent_20():
    return _read(this_dir / "wv20.json", source)


register(
    [
        witherden_vincent_01,
        witherden_vincent_02,
        witherden_vincent_04,
        witherden_vincent_05,
        witherden_vincent_06,
        witherden_vincent_07,
        witherden_vincent_08,
        witherden_vincent_09,
        witherden_vincent_10,
        witherden_vincent_11,
        witherden_vincent_12,
        witherden_vincent_13,
        witherden_vincent_14,
        witherden_vincent_15,
        witherden_vincent_16,
        witherden_vincent_17,
        witherden_vincent_18,
        witherden_vincent_19,
        witherden_vincent_20,
    ]
)



import itertools

import numpy as np


def z(n):
    return np.zeros((1, n), dtype=int)


def rd(n, items):
    """Items is an array of 2-tuples of type (value, count). This method
    returns all all permutations of

    [value1, value1, value2, 0, ..., 0]

    of length n, where value_i occurs count_i times.
    """
    sum_counts = 0
    for item in items:
        _, count = item
        assert count > 0
        sum_counts += count
    assert 0 <= sum_counts <= n

    elems = [((item[0],), item[1]) for item in items]
    elems += [((0,), n - sum_counts)]

    return combine(*elems)


def fsd(n, *tuples):
    """tuples is a list of tuples (value, count). This method returns all
    permutations of [+-r, +-r, +-s, +-s, 0, ..., 0] of length n, with `count`
    times the number `value`.
    """
    pm_tuples = [((+val, -val), count) for val, count in tuples]
    total_count = sum(item[1] for item in tuples)
    assert total_count <= n
    pm_tuples += [((0,), n - total_count)]
    return combine(*pm_tuples)


def fs_array(v):
    """Given an array v = [v0, v1, ..., vn], this returns all permutation of v
    with plus-minus, i.e., [+v0, +v1, ..., +vn], [-v1, +v0, ..., +vn], etc.
    """
    elems = [((+vi, -vi), 1) for vi in v]
    return combine(*elems)


def combine(*elems):
    """Given an input array with lists of options, e.g.,

    ((+r, -r), 2), ((+s, -s), 2), ((0,), 1)

    this methods returns all combinations, i.e.,

    (+r, 0, -s, +r, +s),
    (0, -s, -r, +r, +s),
    ...
    """
    # https://stackoverflow.com/a/45321972/353337

    # Could be replaced by Knuth's "Algorithm L"; see, e.g.,
    # <https://stackoverflow.com/a/4250183/353337>.
    def partitions(*sizes):
        if not sizes or all(s <= 0 for s in sizes):
            yield ()
        for i_size, size in enumerate(sizes):
            if size <= 0:
                continue
            next_sizes = sizes[:i_size] + (sizes[i_size] - 1,) + sizes[i_size + 1 :]
            for p in partitions(*next_sizes):
                yield (i_size,) + p

    values, sizes = zip(*elems)
    templates = partitions(*sizes)
    prod = [itertools.product(*(values[ti] for ti in t)) for t in templates]
    out = np.array(list(itertools.chain.from_iterable(prod)))
    return out


def pm(v):
    """Return all combinations of [+-v[0], ..., +-v[n]] with length n (with repetition).
    The number of combinations depends on the number of zeros.
    """
    possible_vals = []
    for value in v:
        if value == 0:
            possible_vals.append([0])
        else:
            possible_vals.append([+value, -value])

    return np.array(list(itertools.product(*possible_vals)))


def pm_roll(v):
    """Returns `2**k * n` number of points of dimension `n` such that

    p[0] = [+-v[0], ..., +-v[k], 0, ..., 0]
    p[1] = [0, +-v[0], ..., +-v[k], 0, ..., 0]
    ...
    p[n-1] = [+-v[1], ..., +-v[k], 0, ..., 0, +-v[0]]

    with all +- configurations.
    """
    n = len(v)
    pm_v = pm(v)
    return np.concatenate([np.roll(pm_v, i, axis=1) for i in range(n)])


def get_all_exponents(dim, max_degree):
    """Get all exponent combinations of dimension `dim` and maximum degree
    `max_degree`. This method is actually meant for evaluating all polynomials
    with these exponents.

    This problem is similar to the weak_compositions, e.g.,
    <https://stackoverflow.com/a/36748940/353337>. The solution here, however,
    only ever adds 1, making it better suited for a possible extension with
    actual polynomial evaluation.
    """

    def augment(exponents):
        """This function takes the values and exponents of a given monomial
        level, e.g., [(1,0,0), (0,1,0), (0,0,1)], and augments them by one
        level, i.e., [(2,0,0), (1,1,0), (1,0,1), (0,2,0), (0,1,1), (0,0,2)].
        The methods works for all dimension and is based on the observation
        that the augmentation can happen by

          1. adding 1 to all exponents from the previous level (i.e.,
             multiplication with x[0]),
          2. adding 1 to the second exponent of all exponent tuples where
             ex[0]==0 (i.e., multiplication with x[1]),
          3. adding 1 to the third exponent of all exponent tuples where
             ex[0]==0, ex[1]=0 (i.e., multiplication with x[2]),

        etc. The function call is recursive.
        """

        if len(exponents) == 0 or len(exponents[0]) == 0:
            return []

        idx_leading_zero = [k for k in range(len(exponents)) if exponents[k][0] == 0]
        exponents_with_leading_zero = [exponents[k][1:] for k in idx_leading_zero]
        # val1 = vals[idx_leading_zero]
        # x1 = x[1:]
        out1 = augment(exponents_with_leading_zero)

        # increment leading exponent by 1
        out = [[e[0] + 1] + e[1:] for e in exponents]
        # vals0 = vals * x[0]

        out += [[0] + e for e in out1]
        # out_vals = np.concatenate([vals0, vals1])
        return out

    # dim = x.shape[0]

    # Initialization, level 0
    exponents = [dim * [0]]
    # vals = np.array(np.ones(x.shape[1:]))

    # all_vals = []
    all_exponents = []

    # all_vals.append(vals)
    all_exponents.append(exponents)
    for _ in range(max_degree):
        exponents = augment(exponents)
        # all_vals.append(vals)
        all_exponents.append(exponents)

    return all_exponents



import numpy as np

from ..helpers import book, untangle
from ..u3 import _stroud as sphere_stroud
from ._helpers import E3r2Scheme, register
from ._stroud_secrest import stroud_secrest_07 as stroud_e3r2_5_1
from ._stroud_secrest import stroud_secrest_08a as stroud_e3r2_5_2a
from ._stroud_secrest import stroud_secrest_08b as stroud_e3r2_5_2b
from ._stroud_secrest import stroud_secrest_09 as stroud_e3r2_5_3
from ._stroud_secrest import stroud_secrest_10a as stroud_e3r2_7_1a
from ._stroud_secrest import stroud_secrest_10b as stroud_e3r2_7_1b
from ._stroud_secrest import stroud_secrest_11a as stroud_e3r2_7_2a
from ._stroud_secrest import stroud_secrest_11b as stroud_e3r2_7_2b

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_e3r2_14_1(symbolic=False):
    # Get the moments corresponding to monomials and the weight function omega(x) = x^2
    # * exp(-x^2):
    #
    #    int_{-infty}^{infty} x^2 exp(-x^2) x^k dx \
    #
    #           / 0 for k odd,
    #        = {
    #           \ Gamma((k+3)/2) if k even
    #
    # In this particular case, we don't need to compute the recurrence coefficients
    # numerically, but they are given analytically.
    # ```
    # n = 8
    # alpha = np.zeros(n)
    # beta = np.empty(n)
    # beta[0] = np.sqrt(np.pi)/2
    # beta[1::2] = np.arange(n//2) + 1.5
    # beta[2::2] = np.arange(n//2-1) + 1.0
    # points, weights = \
    #     orthopy.line.schemes.custom(alpha, beta, mode='numpy')
    # r = points[-4:]
    # A = weights[-4:]
    # ```
    r = np.array(
        [
            7.235510187528402e-01,
            1.468553289216669e00,
            2.266580584531844e00,
            3.190993201781527e00,
        ]
    )
    A = np.array(
        [
            2.265043732793035e-01,
            1.908084800858996e-01,
            2.539731378612040e-02,
            4.032955750550135e-04,
        ]
    )

    spherical_scheme = sphere_stroud.stroud_u3_14_1()
    v = spherical_scheme.points.T
    B = spherical_scheme.weights

    # Normalize the weights to 1
    B /= np.sqrt(np.pi) / 4

    data = [(A[i] * B[j], r[i] * np.array([v[j]])) for i in range(4) for j in range(72)]
    points, weights = untangle(data)
    d = {"plain": [weights, points[:, 0], points[:, 1], points[:, 2]]}
    return E3r2Scheme("Stroud E3r2 14-1", d, 14, source, 8.903e-14)


register(
    [
        stroud_e3r2_5_1,
        stroud_e3r2_5_2a,
        stroud_e3r2_5_2b,
        stroud_e3r2_5_3,
        stroud_e3r2_7_1a,
        stroud_e3r2_7_1b,
        stroud_e3r2_7_2a,
        stroud_e3r2_7_2b,
        stroud_e3r2_14_1,
    ]
)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C3Scheme, register

_source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Evaluation of Multiple Integrals II",
    journal="Math. Comp.",
    number="12",
    year="1958",
    pages="272-280",
    url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
)


def hammer_stroud_1_3():
    d = {"symm_r00": [[frac(1, 6)], [1]]}
    return C3Scheme("Hammer-Stroud 1-3", d, 3, _source)


def hammer_stroud_2_3():
    alpha = sqrt(frac(3, 5))
    d = {
        "zero3": [[frac(56, 27) / 8]],
        "symm_r00": [[-frac(20, 81) / 8], [alpha]],
        "symm_rr0": [[+frac(50, 81) / 8], [alpha]],
    }
    return C3Scheme("Hammer-Stroud 2-3", d, 5, _source)


def hammer_stroud_4_3():
    d = {
        "symm_r00": [[frac(320, 361) / 8], [sqrt(frac(19, 30))]],
        "symm_rrr": [[frac(121, 361) / 8], [sqrt(frac(19, 33))]],
    }
    return C3Scheme("Hammer-Stroud 4-3", d, 5, _source)


def _hammer_stroud_5_3(variant_a):
    i = 1 if variant_a else -1

    r2 = (33 - i * sqrt(165)) / 28
    s2 = (30 + i * sqrt(165)) / 35
    t2 = (195 - i * 4 * sqrt(165)) / 337

    r = sqrt(r2)
    s = sqrt(s2)
    t = sqrt(t2)

    B1 = 176 / r2 ** 3 / 945
    B2 = 8 / s2 ** 3 / 135
    B3 = 8 / t2 ** 3 / 216
    B0 = 8 - 6 * B1 - 12 * B2 - 8 * B3

    d = {
        "zero3": [[B0 / 8]],
        "symm_r00": [[B1 / 8], [r]],
        "symm_rr0": [[B2 / 8], [s]],
        "symm_rrr": [[B3 / 8], [t]],
    }
    variant = "a" if variant_a else "b"
    return C3Scheme(f"Hammer-Stroud 5-3{variant}", d, 7, _source)


def hammer_stroud_5_3a():
    return _hammer_stroud_5_3(True)


def hammer_stroud_5_3b():
    return _hammer_stroud_5_3(False)


def hammer_stroud_6_3():
    alpha = sqrt(frac(6, 7))
    d = {
        "symm_r00": [[frac(1078, 3645) / 8], [alpha]],
        "symm_rr0": [[frac(343, 3645) / 8], [alpha]],
        "symm_rrr": [
            [0.2247031747656014 / 8, 0.4123338622714356 / 8],
            [0.7341125287521153, 0.4067031864267161],
        ],
    }
    return C3Scheme("Hammer-Stroud 6-3", d, 7, _source)


register(
    [
        hammer_stroud_1_3,
        hammer_stroud_2_3,
        hammer_stroud_4_3,
        hammer_stroud_6_3,
        hammer_stroud_5_3a,
        hammer_stroud_5_3b,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Karin Gatermann"],
    title="Linear Representations of Finite Groups and The Ideal Theoretical Construction of G-Invariant Cubature Formulas",
    journal="Numerical Integration",
    pages="25-35",
    note="Part of the NATO ASI Series book series (ASIC, volume 357)",
)

this_dir = pathlib.Path(__file__).resolve().parent


def gatermann():
    return _read(this_dir / "gatermann.json", source)


register([gatermann])



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, rd, untangle
from ._helpers import TnScheme

source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Integration Over Simplexes",
    journal="Mathematical Tables and Other Aids to Computation",
    volume="10",
    number="55",
    month="jul",
    year="1956",
    pages="137-139",
    url="https://doi.org/10.2307/2002484",
)


def hammer_stroud_1a(n: int):
    degree = 2
    r = (n + 2 - sqrt(n + 2)) / (n + 1) / (n + 2)
    s = (n + 2 + n * sqrt(n + 2)) / (n + 1) / (n + 2)
    data = [(frac(1, n + 1), rd(n + 1, [(r, n), (s, 1)]))]

    points, weights = untangle(data)
    return TnScheme("Hammer-Stround 1a", n, weights, points, degree, source)


def hammer_stroud_1b(n: int):
    degree = 2
    r = (n + 2 + sqrt(n + 2)) / (n + 1) / (n + 2)
    s = (n + 2 - n * sqrt(n + 2)) / (n + 1) / (n + 2)
    data = [(frac(1, n + 1), rd(n + 1, [(r, n), (s, 1)]))]

    points, weights = untangle(data)
    return TnScheme("Hammer-Stround 1b", n, weights, points, degree, source)


def hammer_stroud_2(n: int):
    degree = 3

    B = -frac((n + 1) ** 2, 4 * (n + 2))
    C = frac((n + 3) ** 2, 4 * (n + 1) * (n + 2))

    r = frac(1, n + 1)
    s = frac(1, n + 3)
    t = frac(3, n + 3)

    data = [(B, [(n + 1) * [r]]), (C, rd(n + 1, [(t, 1), (s, n)]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Hammer-Stround 2", n, weights, points, degree, source)



import numpy as np
import orthopy

from ..tools import scheme_from_rc
from ._helpers import C1Scheme


def gauss_legendre(n, mode="numpy"):
    degree = 2 * n - 1

    if mode == "numpy":
        points, weights = np.polynomial.legendre.leggauss(n)
    else:
        rc = orthopy.c1.legendre.RecurrenceCoefficients("monic", symbolic=True)
        _, alpha, beta = np.array([rc[k] for k in range(n)]).T
        points, weights = scheme_from_rc(alpha, beta, rc.int_1, mode=mode)
    return C1Scheme("Gauss-Legendre", degree, weights, points)



def get_good_scheme(degree):
    from ._helpers import schemes

    if degree <= 14:
        return {
            0: schemes["keast_0"],
            1: schemes["keast_0"],
            2: schemes["hammer_marlowe_stroud_1"],
            3: schemes["witherden_vincent_03"],
            4: schemes["liu_vinokur_12"],
            5: schemes["liu_vinokur_12"],
            6: schemes["keast_7"],
            7: schemes["witherden_vincent_07"],
            8: schemes["witherden_vincent_08"],
            9: schemes["witherden_vincent_09"],
            10: schemes["witherden_vincent_10"],
            11: schemes["vioreanu_rokhlin_8"],
            12: schemes["vioreanu_rokhlin_9"],
            13: schemes["vioreanu_rokhlin_9"],
            14: schemes["zhang_cui_liu_2"],
        }[degree]()

    return None



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, untangle, z
from ._helpers import CnScheme, _s2, _s11

_source = article(
    authors=["A.H. Stroud"],
    title="Extensions of Symmetric Integration Formulas",
    journal="Mathematics of Computation",
    volume="22",
    number="102",
    month="apr",
    year="1968",
    pages="271-274",
    url="https://doi.org/10.1090/S0025-5718-1968-0232541-4",
)


def stroud_1968(n):
    r = sqrt(frac(7, 15))
    s, t = (sqrt((7 + i * sqrt(24)) / 15) for i in [+1, -1])
    data = [
        (frac(5 * n ** 2 - 15 * n + 14, 14), z(n)),
        (frac(25, 168), _s2(n, +r)),
        (frac(25, 168), _s2(n, -r)),
        (frac(-25 * (n - 2), 168), fsd(n, (r, 1))),
        (frac(5, 48), _s11(n, +s, -t)),
        (frac(5, 48), _s11(n, -s, +t)),
        (frac(-5 * (n - 2), 48), fsd(n, (s, 1))),
        (frac(-5 * (n - 2), 48), fsd(n, (t, 1))),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme("Stroud 1968", n, weights, points, 5, _source, 2.204e-13)



from sympy import Rational as frac

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["Bernd Meister"],
    title="On a Family of Cubature Formulae",
    journal="Comput J",
    year="1966",
    volume="8",
    number="4",
    pages="368-371",
    url="https://doi.org/10.1093/comjnl/8.4.368",
)


def meister():
    r = frac(2, 3)
    s = frac(1, 3)

    d = {
        "zero2": [[frac(1024, 6720)]],
        "d4_aa": [[frac(576, 6720), -frac(9, 6720), frac(47, 6720)], [r, s, 1]],
        "d4_a0": [[frac(576, 6720)], [r]],
        "d4_ab": [[frac(117, 6720)], [1], [s]],
    }
    return C2Scheme("Meister", d, 7, source)


register([meister])



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["C.R. Morrow", "T.N.L. Patterson"],
    title="The Construction of Algebraic Cubature Formulae by the Distribution of Nodes Along Selected Lines",
    journal="SIAM J. Numer. Anal.",
    volume="22",
    number="6",
    year="1985",
    pages="1178–1190",
    url="https://doi.org/10.1137/0722071",
)

this_dir = pathlib.Path(__file__).resolve().parent


def morrow_patterson_1():
    return _read(this_dir / "morrow_patterson_1.json", source)


def morrow_patterson_2():
    return _read(this_dir / "morrow_patterson_2.json", source)


register([morrow_patterson_1, morrow_patterson_2])



import numpy as np

from ._optimize import optimize


def main():
    import json

    import fjson

    args = _get_parser().parse_args()

    with open(args.infile) as f:
        content = json.load(f)

    name = content["name"]
    try:
        new_data, max_err, cond_in_solution = optimize(content)
    except RuntimeError:
        print(f"{name}: Could not improve scheme any further.")
    else:
        if "weight factor" in content:
            w = content["weight factor"]
            for key, item in new_data.items():
                new_data[key][0] = (np.array(item)[0] / w).tolist()

        if "comments" in content:
            comments = content["comments"]
        else:
            comments = []

        text = "precision improved with quadpy-optimize"
        if text not in comments:
            content["comments"] = comments + [text]

        prev_tol = content["test_tolerance"]
        content["data"] = new_data
        content["test_tolerance"] = max_err

        # make sure that "data" is written last
        keys = list(content.keys())
        if keys[-1] != "data":
            keys.remove("data")
            keys.append("data")
            content = {key: content[key] for key in keys}

        with open(args.infile, "w") as f:
            # Write out 16 digits such that the numbers are preserved exactly when
            # reading the file back it. This makes sure the computed residuals don't
            # change, not even when they are in the range of machine precision.
            fjson.dump(content, f, indent=2, float_format=".16e")
            # for POSIX compliance:
            f.write("\n")
        print(f"{name}:")
        print(f"Improved max error from   {prev_tol}   to   {max_err}.")
        print(f"condition: {cond_in_solution:.3e}")


def _get_parser():
    import argparse
    import sys

    from .__about__ import __version__

    parser = argparse.ArgumentParser(
        description=("Optimize quadrature data."),
        formatter_class=argparse.RawTextHelpFormatter,
    )

    parser.add_argument("infile", type=str, help="quadrature data file to optimize")

    version_text = "\n".join(
        [
            "quadpy {} [Python {}.{}.{}]".format(
                __version__,
                sys.version_info.major,
                sys.version_info.minor,
                sys.version_info.micro,
            ),
            "Copyright (c) 2016-2020 Nico Schlömer",
        ]
    )

    parser.add_argument(
        "--version",
        "-v",
        action="version",
        version=version_text,
        help="display version information",
    )
    return parser



import sympy

from ..helpers import article
from ._helpers import S3Scheme, register

_source = article(
    authors=["V.A. Ditkin"],
    title="On certain approximate formulas for the calculation of triple integrals",
    journal="Doklady Akad. Nauk SSSR (N.S.)",
    number="62",
    year="1948",
    pages="445–447",
    note="Russian",
)

frac = sympy.Rational
sqrt = sympy.sqrt
pi = sympy.pi


def ditkin_1(alpha=0):
    B0 = frac(4, (alpha + 5) ** 2)
    B1 = frac((alpha + 3) * (alpha + 7), 12 * (alpha + 5) ** 2)

    r, s = (sqrt((alpha + 5) * (5 + i * sqrt(5)) / 10 / (alpha + 7)) for i in [+1, -1])

    d = {"zero3": [[B0]], "symm_rs0_roll": [[B1], [r], [s]]}
    return S3Scheme("Ditkin 1", d, 5, _source)


def ditkin_2():
    B0 = frac(4, 25)
    B1 = frac(21, 500)

    r, s = (sqrt((15 + i * 5 * sqrt(5)) / 42) for i in [+1, -1])
    t = sqrt(frac(5, 21))

    d = {
        "zero3": [[B0]],
        "symm_rs0_roll": [[B1], [r], [s]],
        "symm_rrr": [[B1], [t]],
    }
    return S3Scheme("Ditkin 2", d, 5, _source)


def ditkin_3():
    B0 = frac(16, 175)
    B1 = frac(81, 1400)
    B2 = frac(3, 280)

    sqrt5 = sqrt(5)
    r, s = (sqrt((5 + i * sqrt5) / 18) for i in [+1, -1])
    t = sqrt(frac(1, 3))
    u, v = (sqrt((3 - i * sqrt5) / 6) for i in [+1, -1])

    d = {
        "zero3": [[B0]],
        "symm_rs0_roll": [[B1, B2], [r, u], [s, v]],
        "symm_rrr": [[B2], [t]],
    }
    return S3Scheme("Ditkin 3", d, 7, _source)


register([ditkin_1, ditkin_2, ditkin_3])



from sympy import Rational as frac
from sympy import acos, cos, pi, sqrt

from ..helpers import article
from ._helpers import T3Scheme, register

source = article(
    authors=["Y. Liu", "M. Vinokur"],
    title="Exact Integrations of Polynomials and Symmetric Quadrature Formulas over Arbitrary Polyhedral Grids",
    journal="Journal of Computational Physics",
    volume="140",
    pages="122–147",
    year="1998",
    url="https://doi.org/10.1006/jcph.1998.5884",
)


def liu_vinokur_01():
    d = {"s4": [[1]]}
    degree = 1
    return T3Scheme("Liu-Vinokur 1", d, degree, source)


def liu_vinokur_02():
    d = {"s31": [[frac(1, 4)], [0]]}
    degree = 1
    return T3Scheme("Liu-Vinokur 2", d, degree, source)


def liu_vinokur_03():
    d = {"s31": [[frac(1, 4)], [frac(1, 4) - sqrt(5) / 20]]}
    degree = 2
    return T3Scheme("Liu-Vinokur 3", d, degree, source)


def liu_vinokur_04():
    d = {"s4": [[frac(4, 5)]], "s31": [[frac(1, 20)], [0]]}
    degree = 2
    return T3Scheme("Liu-Vinokur 4", d, degree, source)


def liu_vinokur_05():
    d = {"s4": [[-frac(4, 5)]], "s31": [[frac(9, 20)], [frac(1, 6)]]}
    degree = 3
    return T3Scheme("Liu-Vinokur 5", d, degree, source)


def liu_vinokur_06():
    d = {"s31": [[frac(1, 40), frac(9, 40)], [0, frac(1, 3)]]}
    degree = 3
    return T3Scheme("Liu-Vinokur 6", d, degree, source)


def liu_vinokur_07():
    b = (1 + 2 * sqrt(70) / 28) / 4
    d = {
        "s4": [[-frac(148, 1875)]],
        "s31": [[frac(343, 7500)], [frac(1, 14)]],
        "s22": [[frac(56, 375)], [b]],
    }
    degree = 4
    return T3Scheme("Liu-Vinokur 7", d, degree, source)


def liu_vinokur_08():
    alpha1 = (+sqrt(65944 - 19446 * sqrt(11)) + 51 * sqrt(11) - 154) / 89
    alpha2 = (-sqrt(65944 - 19446 * sqrt(11)) + 51 * sqrt(11) - 154) / 89
    d = {
        "s31": [
            [
                (17 * alpha2 - 7) / (420 * alpha1 ** 2 * (alpha2 - alpha1)),
                (17 * alpha1 - 7) / (420 * alpha2 ** 2 * (alpha1 - alpha2)),
            ],
            [(1 - alpha1) / 4, (1 - alpha2) / 4],
        ],
        "s22": [[frac(2, 105)], [0]],
    }
    degree = 4
    return T3Scheme("Liu-Vinokur 8", d, degree, source)


def liu_vinokur_09():
    d = {
        "s4": [[-frac(32, 15)]],
        "s31": [[frac(3, 280), frac(125, 168)], [0, frac(1, 5)]],
        "s22": [[frac(2, 105)], [0]],
    }
    degree = 4
    return T3Scheme("Liu-Vinokur 9", d, degree, source)


def liu_vinokur_10():
    gamma1 = (2 + sqrt(2)) / 4
    gamma2 = (2 - sqrt(2)) / 4
    a = (1 - gamma1 - gamma2) / 4
    b = (1 + 3 * gamma1 - gamma2) / 4
    # c = (1 + 3 * delta - gamma) / 4
    d = {
        "s4": [[frac(32, 105)]],
        "s31": [[-frac(31, 840), frac(27, 280)], [0, frac(1, 3)]],
        "s211": [[frac(4, 105)], [a], [b]],
    }
    degree = 4
    return T3Scheme("Liu-Vinokur 10", d, degree, source)


def liu_vinokur_11():
    d = {
        "s31": [
            [
                (11 - 4 * sqrt(2)) / 840,
                (243 - 108 * sqrt(2)) / 1960,
                (62 + 44 * sqrt(2)) / 735,
            ],
            [0, frac(1, 3), (2 - sqrt(2)) / 4],
        ],
        "s22": [[frac(2, 105)], [0]],
    }
    degree = 4
    return T3Scheme("Liu-Vinokur 11", d, degree, source)


def liu_vinokur_12():
    lmbda = frac(4, 27) * (
        4 * sqrt(79) * cos((acos(67 * sqrt(79) / 24964) + 2 * pi) / 3) + 71
    )
    alpha1 = (+sqrt(9 * lmbda ** 2 - 248 * lmbda + 1680) + 28 - 3 * lmbda) / (
        112 - 10 * lmbda
    )
    alpha2 = (-sqrt(9 * lmbda ** 2 - 248 * lmbda + 1680) + 28 - 3 * lmbda) / (
        112 - 10 * lmbda
    )
    w1 = ((21 - lmbda) * alpha2 - 7) / (420 * alpha1 ** 2 * (alpha2 - alpha1))
    w2 = ((21 - lmbda) * alpha1 - 7) / (420 * alpha2 ** 2 * (alpha1 - alpha2))
    d = {
        "s31": [[w1, w2], [(1 - alpha1) / 4, (1 - alpha2) / 4]],
        "s22": [[lmbda ** 2 / 840], [(1 - 2 / sqrt(lmbda)) / 4]],
    }
    degree = 5
    return T3Scheme("Liu-Vinokur 12", d, degree, source)


def liu_vinokur_13():
    d = {
        "s4": [[-frac(16, 21)]],
        "s31": [
            [(2249 - 391 * sqrt(13)) / 10920, (2249 + 391 * sqrt(13)) / 10920],
            [(1 - (2 + sqrt(13)) / 9) / 4, (1 - (2 - sqrt(13)) / 9) / 4],
        ],
        "s22": [[frac(2, 105)], [0]],
    }
    degree = 5
    return T3Scheme("Liu-Vinokur 13", d, degree, source)


def liu_vinokur_14():
    d = {
        "s4": [[frac(16, 105)]],
        "s31": [
            [frac(1, 280), frac(81, 1400), frac(64, 525)],
            [0, frac(1, 3), frac(1, 8)],
        ],
        "s22": [[frac(2, 105)], [0]],
    }
    degree = 5
    return T3Scheme("Liu-Vinokur 14", d, degree, source)


register(
    [
        liu_vinokur_01,
        liu_vinokur_02,
        liu_vinokur_03,
        liu_vinokur_04,
        liu_vinokur_05,
        liu_vinokur_06,
        liu_vinokur_07,
        liu_vinokur_08,
        liu_vinokur_09,
        liu_vinokur_10,
        liu_vinokur_11,
        liu_vinokur_12,
        liu_vinokur_13,
        liu_vinokur_14,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["P. Bažant", "B.H. Oh"],
    title="Efficient Numerical Integration on the Surface of a Sphere",
    journal="ZAMM",
    volume="66",
    number="1",
    year="1986",
    pages="37-49",
    url="https://doi.org/10.1002/zamm.19860660108",
)

this_dir = pathlib.Path(__file__).resolve().parent


def bazant_oh_09():
    return _read(this_dir / "bazant_oh_09.json", source)


def bazant_oh_11():
    return _read(this_dir / "bazant_oh_11.json", source)


def bazant_oh_13():
    return _read(this_dir / "bazant_oh_13.json", source)


register([bazant_oh_09, bazant_oh_11, bazant_oh_13])



from __future__ import annotations

import math
from typing import Callable

import numpy as np

from .._exception import QuadpyError
from ..helpers import QuadratureScheme


class TnScheme(QuadratureScheme):
    def __init__(
        self,
        name: str,
        dim: int,
        weights,
        points,
        degree: int,
        source,
        tol: float = 1.0e-14,
        comments: list[str] | None = None,
    ):
        self.domain = f"Tn (n={dim})"
        self.dim = dim
        super().__init__(name, weights, points, degree, source, tol, comments)

    def points_inside(self):
        return np.all((0 < self.points) & (self.points < 1))

    def points_inside_or_boundary(self):
        return np.all((0 <= self.points) & (self.points <= 1))

    def integrate(self, f: Callable, simplex, dot=np.dot):
        flt = np.vectorize(float)
        simplex = np.asarray(simplex)
        if simplex.shape[0] != self.dim + 1 or simplex.shape[-1] < self.dim:
            string = ", ".join(str(val) for val in simplex.shape)
            raise QuadpyError(
                f"Wrong domain shape. Expected ({self.dim + 1}, ..., n >= {self.dim}), "
                f"got ({string})."
            )

        x = transform(flt(self.points), simplex.T)
        vol = get_vol(simplex)

        fx = np.asarray(f(x))
        if fx.shape[-len(x.shape[1:]) :] != x.shape[1:]:
            string = ", ".join(str(val) for val in x.shape[1:])
            raise QuadpyError(
                f"Wrong return value shape {fx.shape}. Expected (..., {string})."
            )

        return vol * dot(fx, flt(self.weights))


def transform(points, simplex):
    """Transform the points `xi` from the reference simplex onto `simplex`."""
    # For n == 2:
    # x = (
    #     + outer(triangle[0].T, 1.0 - xi[0] - xi[1])
    #     + outer(triangle[1].T, xi[0])
    #     + outer(triangle[2].T, xi[1])
    #     )
    return np.dot(simplex, points)


def get_vol(simplex):
    # Compute the volume via the Cayley-Menger determinant
    # <http://mathworld.wolfram.com/Cayley-MengerDeterminant.html>. One advantage is
    # that it can compute the volume of the simplex indenpendent of the dimension of the
    # space in which it is embedded.

    # compute all edge lengths
    edges = np.subtract(simplex[:, None], simplex[None, :])
    ei_dot_ej = np.einsum("...k,...k->...", edges, edges)

    j = simplex.shape[0] - 1
    a = np.empty((j + 2, j + 2) + ei_dot_ej.shape[2:])
    a[1:, 1:] = ei_dot_ej
    a[0, 1:] = 1.0
    a[1:, 0] = 1.0
    a[0, 0] = 0.0

    a = np.moveaxis(a, (0, 1), (-2, -1))
    det = np.linalg.det(a)

    vol = np.sqrt((-1.0) ** (j + 1) / 2 ** j / math.factorial(j) ** 2 * det)
    return vol



import pathlib

from sympy import Rational as frac
from sympy import sqrt

from ...helpers import article
from .._helpers import T3Scheme, _read, register

source = article(
    authors=["P. Keast"],
    title="Moderate degree tetrahedral quadrature formulas",
    journal="Computer Methods in Applied Mechanics and Engineering",
    volume="55",
    number="3",
    pages="339-348",
    month="may",
    year="1986",
    url="https://doi.org/10.1016/0045-7825%2886%2990059-9",
)

# https://people.sc.fsu.edu/~jburkardt/datasets/quadrature_rules_tet/quadrature_rules_tet.html
this_dir = pathlib.Path(__file__).resolve().parent


def keast_0():
    # Does not appear in Keast's article. TODO remove
    degree = 1
    d = {"s4": [[1]]}
    return T3Scheme("Keast 0", d, degree, source)


def keast_1():
    # Does not appear in Keast's article.
    degree = 2
    d = {"s31": [[frac(1, 4)], [frac(1, 4) - sqrt(5) / 20]]}
    return T3Scheme("Keast 1", d, degree, source)


def keast_2():
    # Does not appear in Keast's article.
    degree = 3
    d = {"s4": [[-frac(4, 5)]], "s31": [[frac(9, 20)], [frac(1, 6)]]}
    return T3Scheme("Keast 2", d, degree, source)


def keast_3():
    # Does not appear in Keast's article.
    degree = 3
    d = {
        "s31": [[0.2177650698804054], [0.1438564719343852]],
        "s22": [[0.0214899534130631], [0.5]],
    }
    return T3Scheme("Keast 3", d, degree, source)


def keast_4():
    degree = 4
    d = {
        "s4": [[-frac(148, 1875)]],
        "s31": [[frac(343, 7500)], [frac(1, 14)]],
        "s22": [[frac(56, 375)], [frac(1, 4) + sqrt(frac(5, 14)) / 4]],
    }
    return T3Scheme("Keast 4", d, degree, source)


def keast_5():
    degree = 4
    d = {
        "s22": [[2 / 105], [0.5]],
        "s31": [
            [0.0885898247429807, 0.1328387466855907],
            [0.1005267652252045, 0.3143728734931922],
        ],
    }
    return T3Scheme("Keast 5", d, degree, source)


def keast_6():
    degree = 5
    d = {
        "s4": [[frac(6544, 36015)]],
        "s31": [[frac(81, 2240), frac(161051, 2304960)], [frac(1, 3), frac(1, 11)]],
        "s22": [[frac(338, 5145)], [frac(1, 4) - sqrt(91) / 52]],
    }
    return T3Scheme("Keast 6", d, degree, source)


def keast_7():
    return _read(this_dir / "keast_7.json", source)


def keast_8():
    return _read(this_dir / "keast_8.json", source)


def keast_9():
    return _read(this_dir / "keast_9.json", source)


register(
    [
        keast_0,
        keast_1,
        keast_2,
        keast_3,
        keast_4,
        keast_5,
        keast_6,
        keast_7,
        keast_8,
        keast_9,
    ]
)



import numpy as np

from ..helpers import QuadratureScheme, plot_disks_1d


def _find_shapes(fx, intervals, x, domain_shape=None, range_shape=None):
    # This following logic is based on the below requirements
    #
    # assert intervals.shape == (2,) + domain_shape + interval_set_shape
    # assert fx.shape == range_shape + interval_set_shape + x.shape
    #
    assert len(x.shape) == 1

    if intervals.shape[0] != 2:
        raise ValueError(
            f"Expected intervals to bbe of shape (2, ...), "
            f"but got shape {intervals.shape} instead."
        )

    if fx.shape[-1] != x.shape[0]:
        raise ValueError(
            f"Expected the function return value to be of shape (..., {x.shape[0]}), "
            f"but got shape {fx.shape} instead."
        )

    if domain_shape is not None and range_shape is not None:
        # Only needed for some assertions
        interval_set_shape = intervals.shape[1 + len(domain_shape) :]
    elif domain_shape is not None:
        interval_set_shape = intervals.shape[1 + len(domain_shape) :]
        range_shape = fx.shape[: -len(interval_set_shape) - 1]
    elif range_shape is not None:
        interval_set_shape = fx.shape[len(range_shape) : -1]
        if len(interval_set_shape) == 0:
            domain_shape = intervals.shape[1:]
        else:
            domain_shape = intervals.shape[1 : -len(interval_set_shape)]
    else:
        # Find the common tail of fx.shape[:-1] and intervals.shape. That is the
        # interval_set_shape unless the tails of domain_shape and range_shape coincide
        # to some degree. (For this case, one can specify domain_shape, range_shape
        # explicitly.)
        interval_set_shape = []
        for k in range(min(len(intervals.shape) - 1, len(fx.shape) - 1)):
            d = intervals.shape[-k - 1]
            if fx.shape[-k - 2] != d:
                break
            interval_set_shape.append(d)
        interval_set_shape = tuple(reversed(interval_set_shape))

        if interval_set_shape == ():
            domain_shape = intervals.shape[1:]
        else:
            domain_shape = intervals.shape[1 : -len(interval_set_shape)]
        range_shape = fx.shape[: -len(interval_set_shape) - 1]

    expected_shape_ivals = (2,) + domain_shape + interval_set_shape
    if intervals.shape != expected_shape_ivals:
        raise ValueError(
            f"Expected intervals to be of shape {expected_shape_ivals}, "
            f"but got shape {intervals.shape} instead."
        )
    expected_shape_fx = range_shape + interval_set_shape + x.shape
    if fx.shape != expected_shape_fx:
        raise ValueError(
            f"Expected the function return value to be of shape {expected_shape_fx}, "
            f"but got shape {fx.shape} instead."
        )
    return domain_shape, range_shape, interval_set_shape


class C1Scheme(QuadratureScheme):
    def __init__(self, name, degree, weights, points, source=None):
        self.domain = "C1"
        super().__init__(name, weights, points, degree, source)

    def integrate(self, f, intervals, domain_shape=None, range_shape=None, dot=np.dot):
        iv = np.asarray(intervals)
        x0 = 0.5 * (1.0 - self.points)
        x1 = 0.5 * (1.0 + self.points)
        x = np.multiply.outer(iv[0], x0) + np.multiply.outer(iv[1], x1)
        fx = np.asarray(f(x))

        domain_shape, range_shape, _ = _find_shapes(
            fx, iv, self.points, domain_shape=domain_shape, range_shape=range_shape
        )

        # np.sum is slower than dot() and friends, but allows for scalar input.
        diff = iv[1] - iv[0]
        len_intervals = np.sqrt(np.sum(diff ** 2, axis=tuple(range(len(domain_shape)))))
        # The factor 0.5 is from the length of the reference line [-1, 1].
        return 0.5 * len_intervals * dot(fx, self.weights)

    def plot(self, interval=np.array([[-1.0], [1.0]]), show_axes=False):
        from matplotlib import pyplot as plt

        # change default range so that new disks will work
        plt.axis("equal")
        # ax.set_xlim((-1.5, 1.5))
        # ax.set_ylim((-1.5, 1.5))

        if not show_axes:
            plt.gca().set_axis_off()

        plt.plot(interval, [0, 0], color="k")

        pts = np.column_stack([self.points, np.zeros(len(self.points))])
        total_area = interval[1] - interval[0]
        plot_disks_1d(plt, pts, self.weights, total_area)



import numpy as np
from sympy import Rational as frac

from ..helpers import article, rd, untangle
from ._helpers import TnScheme

source = article(
    authors=["R. Lauffer"],
    title="Interpolation mehfacher Integrale",
    journal="Arch. Math.",
    volume="6",
    year="1955",
    pages="159-164",
    url="https://doi.org/10.1007/BF01900222",
)


def lauffer_1(n: int):
    data = [(frac(1, n + 1), rd(n + 1, [(1, 1)]))]
    points, weights = untangle(data)
    return TnScheme("Lauffer 1", n, weights, points, 1, source)


def lauffer_2(n: int):
    B = frac(2 - n, (n + 1) * (n + 2))
    C = frac(4, (n + 1) * (n + 2))
    data = [(B, rd(n + 1, [(1, 1)])), (C, rd(n + 1, [(frac(1, 2), 2)]))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Lauffer 2", n, weights, points, 2, source)


def lauffer_3(n: int):
    B = frac(n ** 2 - 4 * n + 6, (n + 1) * (n + 2) * (n + 3))
    C = frac(27 - 9 * n, 2 * (n + 1) * (n + 2) * (n + 3))
    D = frac(27, (n + 1) * (n + 2) * (n + 3))

    r = frac(1, 3)
    s = frac(2, 3)

    data = [
        (B, rd(n + 1, [(1, 1)])),
        (C, rd(n + 1, [(r, 1), (s, 1)])),
        (D, rd(n + 1, [(r, 3)])),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Lauffer 3", n, weights, points, 3, source)


def lauffer_4(n: int):
    assert n >= 3

    nprod = (n + 1) * (n + 2) * (n + 3) * (n + 4)
    B1 = frac(-3 * n ** 3 + 17 * n ** 2 - 58 * n + 72, 3 * nprod)
    B2 = frac(16 * (n ** 2 - 5 * n + 12), 3 * nprod)
    B3 = frac(4 * (n ** 2 - 9 * n + 12), nprod)
    B4 = frac(64 * (4 - n), 2 * nprod)
    B5 = frac(256, nprod)

    r = frac(1, 4)
    s = frac(3, 4)
    t = frac(1, 2)

    data = [
        (B1, rd(n + 1, [(1, 1)])),
        (B2, rd(n + 1, [(r, 1), (s, 1)])),
        (B3, rd(n + 1, [(t, 2)])),
        (B4, rd(n + 1, [(r, 2), (t, 1)])),
        (B5, rd(n + 1, [(r, 4)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Lauffer 4", n, weights, points, 4, source)


def lauffer_5(n: int):
    assert n >= 4

    nprod = (n + 1) * (n + 2) * (n + 3) * (n + 4) * (n + 5)

    # ERR Stroud is missing the factor 1/12 in B1.
    B1 = frac((12 * n ** 4 - 82 * n ** 3 + 477 * n ** 2 - 1277 * n + 1440), 12 * nprod)
    B2 = frac(5 ** 2 * (-3 * n ** 3 + 19 * n ** 2 - 96 * n + 170), 12 * nprod)
    B3 = frac(5 ** 2 * (-(n ** 3) + 13 * n ** 2 - 47 * n + 65), 6 * nprod)
    B4 = frac(5 ** 3 * (n ** 2 - 6 * n + 20), 3 * nprod)
    B5 = frac(5 ** 3 * (n ** 2 - 11 * n + 20), 4 * nprod)
    B6 = frac(5 ** 4 * (5 - n), 2 * nprod)
    B7 = frac(5 ** 5, nprod)

    r = frac(1, 5)
    s = frac(4, 5)
    u = frac(2, 5)
    v = frac(3, 5)

    data = [
        (B1, rd(n + 1, [(1, 1)])),
        (B2, rd(n + 1, [(r, 1), (s, 1)])),
        (B3, rd(n + 1, [(u, 1), (v, 1)])),
        (B4, rd(n + 1, [(r, 2), (v, 1)])),
        (B5, rd(n + 1, [(r, 1), (u, 2)])),
        (B6, rd(n + 1, [(r, 3), (u, 1)])),
        (B7, rd(n + 1, [(r, 5)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Lauffer 5", n, weights, points, 5, source)



import numpy as np

from ..helpers import article
from ._helpers import S2Scheme, register

_source = article(
    authors=["Frank G. Lether"],
    title="A Generalized Product Rule for the Circle",
    journal="SIAM Journal on Numerical Analysis",
    volume="8",
    number="2",
    month="jun",
    year="1971",
    pages="249-253",
    url="https://www.jstor.org/stable/2949473",
)


def lether(n: int):
    assert n >= 1
    p, w = np.polynomial.legendre.leggauss(n)

    mu = np.arange(1, n + 1)
    points = np.array(
        [
            np.tile(np.cos(mu * np.pi / (n + 1)), n),
            np.outer(p, np.sin(mu * np.pi / (n + 1))).flatten(),
        ]
    )
    weights = np.outer(w, np.sin(mu * np.pi / (n + 1)) ** 2).flatten() / (n + 1)
    d = {"plain": [weights, points[0], points[1]]}
    return S2Scheme(f"Lether({n})", d, 2 * n - 1, _source)


register([lether])



import pathlib

from ...helpers import online
from .._helpers import _read, register

source = online(
    authors=["Jörg Fliege", "Ulrike Maier"],
    title="A Two-Stage Approach for Computing Cubature Formulae for the Sphere",
    url="https://www.personal.soton.ac.uk/jf1w07/nodes/nodes.html",
)

this_dir = pathlib.Path(__file__).resolve().parent


def fliege_maier_04():
    return _read(this_dir / "fliege_maier_04.json", source)


def fliege_maier_09():
    return _read(this_dir / "fliege_maier_09.json", source)


def fliege_maier_16():
    return _read(this_dir / "fliege_maier_16.json", source)


def fliege_maier_25():
    return _read(this_dir / "fliege_maier_25.json", source)


register([fliege_maier_04, fliege_maier_09, fliege_maier_16, fliege_maier_25])



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["J.I. Maeztu", "E. Sainz de la Maza"],
    title="An invariant quadrature rule of degree 11 for the tetrahedron",
    journal="C. R. Acad. Sci. Paris",
    volume="321",
    year="1995",
    pages="1263-1267",
)

this_dir = pathlib.Path(__file__).resolve().parent


def maeztu_sainz():
    # The article claims degree 11, but tests show only degree 1. :/
    # TODO find out what's going on
    return _read(this_dir / "maeztu_sainz.json", source)


register([maeztu_sainz])



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import book
from ._helpers import E2rScheme, register

# ERR misprint in Stroud copied from original article; rabinowitz_richter_4 as stroud_13_1,
from ._rabinowitz_richter import rabinowitz_richter_1 as stroud_9_1
from ._rabinowitz_richter import rabinowitz_richter_2 as stroud_11_1
from ._rabinowitz_richter import rabinowitz_richter_3 as stroud_11_2
from ._rabinowitz_richter import rabinowitz_richter_5 as stroud_15_1
from ._stroud_secrest import stroud_secrest_5 as stroud_5_1
from ._stroud_secrest import stroud_secrest_6 as stroud_7_1

_source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_4_1():
    d = {
        "zero2": [[frac(7, 10)]],
        "d5.0": [[frac(3, 50)], [[2 * sqrt(5)]]],
    }
    return E2rScheme("Stroud 4-1", d, 4, _source)


register(
    [
        stroud_4_1,
        stroud_5_1,
        stroud_7_1,
        stroud_9_1,
        stroud_11_1,
        stroud_11_2,
        stroud_15_1,
    ]
)



from . import (
    _albrecht,
    _albrecht_collatz,
    _classical,
    _cools_haegemans,
    _cools_kim,
    _haegemans_piessens,
    _hammer_stroud,
    _kim_song,
    _lether,
    _luo_meng,
    _mysovskih,
    _peirce_1956,
    _peirce_1957,
    _piessens_haegemans,
    _rabinowitz_richter,
    _radon,
    _stroud,
    _wissmann_becker,
)
from ._helpers import get_good_scheme, schemes

__all__ = [
    "_albrecht",
    "_albrecht_collatz",
    "_classical",
    "_cools_haegemans",
    "_cools_kim",
    "_haegemans_piessens",
    "_hammer_stroud",
    "_kim_song",
    "_lether",
    "_luo_meng",
    "_mysovskih",
    "_peirce_1956",
    "_peirce_1957",
    "_piessens_haegemans",
    "_rabinowitz_richter",
    "_radon",
    "_stroud",
    "_wissmann_becker",
    #
    "schemes",
    "get_good_scheme",
]



import pathlib

from sympy import Rational as frac
from sympy import sqrt

from ...helpers import article
from .._helpers import C2Scheme, _read, register

source = article(
    authors=["H.J. Schmid"],
    title="On cubature formulae with a minimal number of knots",
    journal="Numerische Mathematik",
    month="sep",
    year="1978",
    volume="31",
    number="3",
    pages="281–297",
    url="https://eudml.org/doc/132580",
)

this_dir = pathlib.Path(__file__).resolve().parent


def schmid_2():
    d = {
        "plain": [
            [frac(1, 4), frac(1, 4), frac(1, 2)],
            [-sqrt(frac(1, 3)), -sqrt(frac(1, 3)), +sqrt(frac(1, 3))],
            [+sqrt(frac(2, 3)), -sqrt(frac(2, 3)), 0],
        ]
    }
    return C2Scheme("Schmid 2", d, 2, source, 4.441e-16)


def schmid_4():
    d = {
        "plain": [
            [
                frac(2, 9) - 2 * sqrt(5) / 45,
                frac(2, 9) + 2 * sqrt(5) / 45,
                frac(5, 36) + 5 * sqrt(29) / 18 / 29,
                frac(5, 36) + 5 * sqrt(29) / 18 / 29,
                frac(5, 36) - 5 * sqrt(29) / 18 / 29,
                frac(5, 36) - 5 * sqrt(29) / 18 / 29,
            ],
            [0, 0, +sqrt(15) / 5, -sqrt(15) / 5, +sqrt(15) / 5, -sqrt(15) / 5],
            [
                (sqrt(3) + sqrt(15)) / 6,
                (sqrt(3) - sqrt(15)) / 6,
                (+sqrt(87) - 2 * sqrt(3)) / 15,
                (+sqrt(87) - 2 * sqrt(3)) / 15,
                (-sqrt(87) - 2 * sqrt(3)) / 15,
                (-sqrt(87) - 2 * sqrt(3)) / 15,
            ],
        ]
    }
    return C2Scheme("Schmid 4", d, 4, source, 4.441e-16)


def schmid_6():
    return _read(this_dir / "schmid_6.json", source)


register([schmid_2, schmid_4, schmid_6])



import math

import numpy as np
import sympy

from ..cn._cools_haegemans import _gener
from ..helpers import article, fsd, pm, untangle, z
from ._helpers import Enr2Scheme

_source = article(
    authors=["Ronald Cools", "Ann Haegemans"],
    title="An imbedded family of cubature formulae for n-dimensional product regions",
    journal="Journal of Computational and Applied Mathematics",
    volume="51",
    year="1994",
    pages="251-262",
    url="https://doi.org/10.1016/0377-0427%2892%2900007-V",
)


def _mu(j, symbolic):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    # 1/sqrt(pi) ** n int int ... int exp(-r ** 2) dr
    if j == 0:
        return 1
    elif j == 1:
        return 0
    return frac(j - 1, 2) * _mu(j - 2, symbolic)


def cools_haegemans_1(n, delta2=1, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert frac(1, 2) <= delta2
    m = 1

    w0 = frac(2 * delta2 - 1, 2 * delta2)
    w = frac(_mu(2, symbolic) ** m * _mu(0, symbolic) ** (n - m), 2 ** n * delta2 ** m)

    data = [
        (w0, z(n)),
        (w, pm(n * [sqrt(delta2)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Cools-Haegemans 1", n, weights, points, 3, _source)


def cools_haegemans_2(n, delta2=1, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    assert n >= 1
    assert frac(1, 2) <= delta2
    if n > 2:
        assert delta2 <= frac(n + 2, 2 * n - 4)
    m = 2

    lmbdas2 = _gener(delta2, 1, _mu, symbolic)

    w0 = frac(
        -(2 * delta2 - 1) * (2 * delta2 * n - 4 * delta2 - n - 2), 8 * delta2 ** 2
    )
    w1 = frac((2 * delta2 - 1) ** 2, 16 * delta2 ** 2)
    w = frac(_mu(2, symbolic) ** m * _mu(0, symbolic) ** (n - m), 2 ** n * delta2 ** m)

    lmbdas = [sqrt(lmbda2) for lmbda2 in lmbdas2]

    data = [
        (w0, z(n)),
        (w1, fsd(n, (lmbdas[0], 1))),
        (w, pm(n * [sqrt(delta2)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Cools-Haegemans 2", n, weights, points, 5, _source)


def cools_haegemans_3(n, delta2=(2, 3), symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    if isinstance(delta2, tuple):
        delta2 = frac(*delta2)

    assert n >= 2
    m = 3

    lmbdas2 = _gener(delta2, 2, _mu, symbolic)

    delta4 = delta2 ** 2
    delta6 = delta2 ** 3

    w0 = frac(
        (
            delta6 * (16 * n ** 2 - 72 * n + 128)
            + delta4 * (-28 * n ** 2 + 48 * n - 32)
            + delta2 * (16 * n ** 2 + 30 * n - 16)
            + (-3 * n ** 2 - 18 * n - 24)
        )
        * (2 * delta2 - 1),
        64 * (4 * delta2 - 3) * delta6,
    )
    w1 = -frac(
        (12 * delta4 * n - 32 * delta4 - 12 * delta2 * n + 3 * n + 12)
        * (2 * delta2 - 1),
        192 * delta6,
    )
    w2 = frac((2 * delta2 - 1) * (2 * delta2 - 3) ** 3, 384 * (4 * delta2 - 3) * delta6)
    w11 = frac((2 * delta2 - 1) ** 3, 128 * delta6)

    w = frac(_mu(2, symbolic) ** m * _mu(0, symbolic) ** (n - m), 2 ** n * delta2 ** m)

    delta = sqrt(delta2)
    lmbdas = [sqrt(lmbda2) for lmbda2 in lmbdas2]

    data = [
        (w0, z(n)),
        (w1, fsd(n, (lmbdas[0], 1))),
        (w2, fsd(n, (lmbdas[1], 1))),
        (w11, fsd(n, (lmbdas[0], 2))),
        (w, pm(n * [delta])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Cools-Haegemans 3", n, weights, points, 7, _source)


# ERR There is a mistake here somewhere in the weights, but it's unclear where.
# TODO fix this
# def cools_haegemans_4(n, delta2=frac(2, 3)):
#     assert n >= 2
#     m = 4
#
#     lmbdas2 = _gener(delta2, 3, _mu)
#
#     delta4 = delta2 ** 2
#     delta6 = delta2 ** 3
#     delta8 = delta2 ** 4
#
#     w2 = -frac(
#         (
#             delta8 * (48 * n - 160)
#             + delta6 * (192 * n - 480)
#             + delta4 * (-408 * n + 96)
#             + delta2 * (240 * n + 600)
#             + (-45 * n - 270)
#         )
#         * (2 * delta2 - 3) ** 3,
#         4608 * (4 * delta4 + 20 * delta2 - 15) * (4 * delta2 - 3) * delta8,
#     )
#     w11 = -frac(
#         (12 * delta4 * n - 40 * delta4 - 12 * delta2 * n + 3 * n + 18)
#         * (2 * delta2 - 1) ** 2,
#         1536 * delta8,
#     )
#     w3 = frac(
#         (4 * delta4 - 24 * delta2 + 15) ** 4 * (2 * delta2 - 3) ** 3,
#         4608
#         * (8 * delta6 - 324 * delta4 + 414 * delta2 - 135)
#         * (4 * delta4 + 20 * delta2 - 15)
#         * (22 * delta2 - 15)
#         * delta8,
#     )
#     w21 = frac(
#         (2 * delta2 - 1) ** 2 * (2 * delta2 - 3) ** 3, 3072 * (4 * delta2 - 3) * delta8
#     )
#     w111 = frac((2 * delta2 - 1) ** 4, 1024 * delta8)
#
#     w = frac(_mu(2) ** m * _mu(0) ** (n - m), 2 ** n * delta2 ** m)
#
#     delta = sqrt(delta2)
#     lmbdas = [sqrt(lmbda2) for lmbda2 in lmbdas2]
#
#     data = [
#         (w2, fsd(n, (lmbdas[1], 1))),
#         (w11, fsd(n, (lmbdas[0], 2))),
#         (w3, fsd(n, (lmbdas[2], 1))),
#         (w21, fsd(n, (lmbdas[1], 1), (lmbdas[0], 1))),
#         (w111, fsd(n, (lmbdas[0], 3))),
#         (w, pm(n, delta)),
#     ]
#
#     # This is an attempt to find the correct symmetries, but to no avail. Something
#     # appears to be wrong with the weights.
#     print(n)
#     maxk = 30
#     k = 2 ** n
#     print("try")
#     for k2 in range(0, maxk, 2):
#         for k11 in range(0, maxk, 2):
#             for k3 in range(0, maxk, 2):
#                 for k21 in range(0, maxk, 2):
#                     for k111 in range(0, maxk, 2):
#                         val = (
#                             k2 * w2
#                             + k11 * w11
#                             + k3 * w3
#                             + k21 * w21
#                             + k111 * w111
#                             + k * w
#                         )
#                         if val == 1:
#                             print("SUCCESS", k2, k11, k3, k21, k111, k)
#     exit(1)
#
#     points, weights = untangle(data)
#
#     print(points)
#     print(weights)
#     print(sum(weights))
#     exit(1)
#     return Enr2Scheme("Cools-Haegemans 4", n, weights, points, 9, _source)



import math

import numpy as np
import sympy

from ..helpers import book, pm, untangle
from ._hammer_stroud import hammer_stroud_11n, hammer_stroud_12n
from ._helpers import SnScheme
from ._stenger import stenger_7a as stroud_sn_7_3a
from ._stenger import stenger_7b as stroud_sn_7_3b
from ._stenger import stenger_9a as stroud_sn_9_1a
from ._stenger import stenger_9b as stroud_sn_9_1b
from ._stenger import stenger_11a as stroud_sn_11_1a
from ._stenger import stenger_11b as stroud_sn_11_1b
from ._stroud_1957 import stroud_1957 as stroud_sn_2_1
from ._stroud_1966 import stroud_1966_a as stroud_sn_5_3
from ._stroud_1966 import stroud_1966_b as stroud_sn_5_4
from ._stroud_1966 import stroud_1966_c as stroud_sn_5_5
from ._stroud_1966 import stroud_1966_d as stroud_sn_5_6
from ._stroud_1967_5 import stroud_1967_5_a as stroud_sn_5_1a
from ._stroud_1967_5 import stroud_1967_5_b as stroud_sn_5_1b
from ._stroud_1967_7 import stroud_1967_7_a as stroud_sn_7_1a
from ._stroud_1967_7 import stroud_1967_7_b as stroud_sn_7_1b
from ._stroud_1967_7 import stroud_1967_7_c as stroud_sn_7_2

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_sn_3_1(n):
    return hammer_stroud_11n(n, 0)


def stroud_sn_3_2(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    r = sqrt(frac(1, n + 2))
    data = [(frac(1, 2 ** n), pm(n * [r]))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stroud Sn 3-2", n, weights, points, 3, source)


def stroud_sn_5_2(n):
    return hammer_stroud_12n(n, 0)


__all__ = [
    "stroud_sn_2_1",
    "stroud_sn_3_1",
    "stroud_sn_3_2",
    "stroud_sn_5_1a",
    "stroud_sn_5_1b",
    "stroud_sn_5_2",
    "stroud_sn_5_3",
    "stroud_sn_5_4",
    "stroud_sn_5_5",
    "stroud_sn_5_6",
    "stroud_sn_7_1a",
    "stroud_sn_7_1b",
    "stroud_sn_7_2",
    "stroud_sn_7_3a",
    "stroud_sn_7_3b",
    "stroud_sn_9_1a",
    "stroud_sn_9_1b",
    "stroud_sn_11_1a",
    "stroud_sn_11_1b",
]



from sympy import Rational as frac

from ._helpers import C3Scheme, register


def midpoint():
    d = {"zero3": [[1]]}
    return C3Scheme("Midpoint", d, 1)


def vertex():
    d = {"symm_rrr": [[frac(1, 8)], [1]]}
    return C3Scheme("Vertex", d, 1)


def face_midpoint():
    d = {"symm_r00": [[frac(1, 6)], [1]]}
    return C3Scheme("Face-Midpoint", d, 3)


register([midpoint, face_midpoint, vertex])



from sympy import Rational as frac

from ..helpers import article
from ._helpers import C3Scheme, register

source = article(
    authors=["G.W. Tyler"],
    title="Numerical integration of functions of several variables",
    journal="Canad. J. Math.",
    volume="5",
    year="1953",
    pages="393-412",
    url="https://doi.org/10.4153/CJM-1953-044-1",
)


def tyler_1():
    d = {"symm_r00": [[frac(1, 6)], [1]]}
    return C3Scheme("Tyler 1", d, 3, source)


def tyler_2():
    d = {
        "zero3": [[-frac(62, 45)]],
        "symm_r00": [[frac(16, 45), frac(1, 45)], [frac(1, 2), 1]],
        "symm_rrr": [[frac(1, 72)], [1]],
    }
    return C3Scheme("Tyler 2", d, 5, source)


register([tyler_1, tyler_2])



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import E3r2Scheme, register

source = article(
    authors=["A.H. Stroud", "D. Secrest"],
    title="Approximate integration formulas for certain spherically symmetric regions",
    journal="Math. Comp.",
    volume="17",
    year="1963",
    pages="105-135",
    url="https://doi.org/10.1090/S0025-5718-1963-0161473-0",
)


def stroud_secrest_07():
    # ERR
    # article:
    # nu, xi = sqrt((15 + plus_minus * 3*sqrt(5)))
    # A = 3/5
    # B = 1/30

    # book:
    nu, xi = (sqrt((5 - p_m * sqrt(5)) / 4) for p_m in [+1, -1])
    A = frac(2, 5)
    B = frac(1, 20)

    d = {"zero3": [[A]], "symm_rs0_roll": [[B], [nu], [xi]]}
    return E3r2Scheme("Stroud-Secrest VII", d, 5, source)


def stroud_secrest_08a():
    r = sqrt(frac(5, 4))
    s = sqrt(frac(5, 2))
    d = {
        "symm_r00": [[frac(4, 25)], [r]],
        "symm_rrr": [[frac(1, 200)], [s]],
    }
    return E3r2Scheme("Stroud-Secrest VIIIa", d, 5, source)


def stroud_secrest_08b():
    r = sqrt(frac(5, 2))
    s = sqrt(frac(5, 6))
    d = {
        "zero3": [[frac(2, 5)]],
        "symm_r00": [[frac(1, 25)], [r]],
        "symm_rrr": [[frac(9, 200)], [s]],
    }
    return E3r2Scheme("Stroud-Secrest VIIIb", d, 5, source)


def stroud_secrest_09():
    r, s = (sqrt((15 - p_m * 5 * sqrt(5)) / 12) for p_m in [+1, -1])
    t = sqrt(frac(5, 6))
    d = {
        "zero3": [[frac(2, 5)]],
        "symm_rs0_roll": [[frac(3, 100)], [r], [s]],
        "symm_rrr": [[frac(3, 100)], [t]],
    }
    return E3r2Scheme("Stroud-Secrest IX", d, 5, source)


def _stroud_secrest_10(positive):
    plus_minus = 1 if positive else -1

    sqrt15 = sqrt(15)

    r = sqrt((15 + plus_minus * sqrt15) / 4)
    s = sqrt((6 - plus_minus * sqrt15) / 2)
    t = sqrt((9 + plus_minus * 2 * sqrt15) / 2)
    A = (720 + plus_minus * 8 * sqrt15) / 2205
    B = (270 - plus_minus * 46 * sqrt15) / 15435
    C = (162 + plus_minus * 41 * sqrt15) / 6174
    D = (783 - plus_minus * 202 * sqrt15) / 24696

    d = {
        "zero3": [[A]],
        "symm_r00": [[B], [r]],
        "symm_rr0": [[C], [s]],
        "symm_rrr": [[D], [t]],
    }
    variant = "a" if positive else "b"
    return E3r2Scheme(f"Stroud-Secrest X{variant}", d, 7, source)


def stroud_secrest_10a():
    return _stroud_secrest_10(True)


def stroud_secrest_10b():
    return _stroud_secrest_10(False)


def _stroud_secrest_11(positive):
    p_m = 1 if positive else -1

    sqrt2 = sqrt(2)
    sqrt5 = sqrt(5)
    sqrt10 = sqrt(10)

    r = sqrt((25 + p_m * 15 * sqrt2 + 5 * sqrt5 + p_m * 3 * sqrt10) / 4)
    s = sqrt((25 + p_m * 15 * sqrt2 - 5 * sqrt5 - p_m * 3 * sqrt10) / 4)
    t = sqrt((3 - p_m * sqrt2) / 2)
    u = sqrt((9 - p_m * 3 * sqrt2 - 3 * sqrt5 + p_m * sqrt10) / 4)
    v = sqrt((9 - p_m * 3 * sqrt2 + 3 * sqrt5 - p_m * sqrt10) / 4)

    A = (80 + p_m * 8 * sqrt2) / 245
    B = (395 - p_m * 279 * sqrt2) / 13720
    C = (45 + p_m * 29 * sqrt2) / 2744

    d = {
        "zero3": [[A]],
        "symm_rs0_roll": [[B, C], [r, u], [s, v]],
        "symm_rrr": [[C], [t]],
    }
    return E3r2Scheme("Stroud-Secrest XI", d, 7, source)


def stroud_secrest_11a():
    return _stroud_secrest_11(True)


def stroud_secrest_11b():
    return _stroud_secrest_11(False)


register(
    [
        stroud_secrest_07,
        stroud_secrest_09,
        stroud_secrest_08a,
        stroud_secrest_08b,
        stroud_secrest_10a,
        stroud_secrest_10b,
        stroud_secrest_11a,
        stroud_secrest_11b,
    ]
)



from . import (
    _cools_haegemans,
    _haegemans_piessens,
    _rabinowitz_richter,
    _stroud,
    _stroud_secrest,
)
from ._helpers import get_good_scheme, schemes

__all__ = [
    "_cools_haegemans",
    "_haegemans_piessens",
    "_rabinowitz_richter",
    "_stroud",
    "_stroud_secrest",
    #
    "schemes",
    "get_good_scheme",
]



import sympy

from ..helpers import article
from ._helpers import S3Scheme, register

_source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Evaluation of Multiple Integrals II",
    journal="Math. Comp.",
    number="12",
    year="1958",
    pages="272-280",
    url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
)

frac = sympy.Rational
sqrt = sympy.sqrt
pi = sympy.pi


def hammer_stroud_11_3():
    d = {"symm_r00": [[frac(1, 6)], [sqrt(frac(3, 5))]]}
    return S3Scheme("Hammer-Stroud 11-3", d, 3, _source)


def hammer_stroud_12_3():
    alpha = sqrt(frac(3, 7))
    d = {
        "zero3": [[frac(1, 15)]],
        "symm_r00": [[frac(7, 90)], [alpha]],
        "symm_rr0": [[frac(7, 180)], [alpha]],
    }
    return S3Scheme("Hammer-Stroud 12-3", d, 5, _source)


def hammer_stroud_14_3(variant_a=True):
    t = 1 if variant_a else -1

    sqrt14 = sqrt(14)

    # ERR The article incorrectly gives 0.50824... instead of 0.050824...
    a1 = frac(1, 125) * (9 + t * 2 * sqrt14)
    c1 = (71 - t * 12 * sqrt14) / 1000

    nu = sqrt((7 - t * sqrt14) / 7)
    eta1 = sqrt(5 / (21 - t * 2 * sqrt14))

    d = {"symm_r00": [[a1], [nu]], "symm_rrr": [[c1], [eta1]]}
    name = "Hammer-Stroud 14-3" + ("a" if variant_a else "b")
    return S3Scheme(name, d, 5, _source)


def _hammer_stroud_15_3(variant_a):
    t = 1 if variant_a else -1

    sqrt30 = sqrt(30)
    nu2 = (45 - t * sqrt30) / 57
    xi2 = (18 + t * sqrt30) / 42
    eta2 = 7 / (27 + t * 2 * sqrt30)

    # The extract expressions are from Stroud's book.
    a1 = 1 / nu2 ** 3 / 63
    b1 = 1 / xi2 ** 3 / 630
    c1 = 1 / eta2 ** 3 / 2520
    a0 = 1 - 6 * a1 - 12 * b1 - 8 * c1

    d = {
        "zero3": [[a0]],
        "symm_r00": [[a1], [sqrt(nu2)]],
        "symm_rr0": [[b1], [sqrt(xi2)]],
        "symm_rrr": [[c1], [sqrt(eta2)]],
    }
    name = "Hammer-Stroud 15-3" + ("a" if variant_a else "b")
    return S3Scheme(name, d, 7, _source)


def hammer_stroud_15_3a():
    return _hammer_stroud_15_3(True)


def hammer_stroud_15_3b():
    return _hammer_stroud_15_3(False)


register(
    [
        hammer_stroud_11_3,
        hammer_stroud_12_3,
        hammer_stroud_14_3,
        hammer_stroud_15_3a,
        hammer_stroud_15_3b,
    ]
)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import book, fsd, pm, untangle
from ._helpers import UnScheme
from ._stroud_1967 import stroud_1967 as stroud_un_7_1
from ._stroud_1969 import stroud_1969 as stroud_un_11_1

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_un_3_1(n):
    degree = 3
    data = [(frac(1, 2 * n), fsd(n, (1, 1)))]
    points, weights = untangle(data)
    return UnScheme("Stroud Un 3-1", n, weights, points, degree, source)


def stroud_un_3_2(n):
    degree = 3
    data = [(frac(1, 2 ** n), pm(n * [sqrt(frac(1, n))]))]
    points, weights = untangle(data)
    return UnScheme("Stroud Un 3-2", n, weights, points, degree, source)


def stroud_un_5_1(n):
    degree = 5

    B1 = frac(4 - n, 2 * n * (n + 2))
    B2 = frac(1, n * (n + 2))

    data = [(B1, fsd(n, (1, 1))), (B2, fsd(n, (sqrt(frac(1, 2)), 2)))]

    points, weights = untangle(data)
    return UnScheme("Stroud Un 5-1", n, weights, points, degree, source)


def stroud_un_5_2(n):
    degree = 5

    B1 = frac(1, n * (n + 2))
    B2 = frac(n, 2 ** n * (n + 2))

    data = [(B1, fsd(n, (1, 1))), (B2, pm(n * [sqrt(frac(1, n))]))]

    points, weights = untangle(data)
    return UnScheme("Stroud Un 5-2", n, weights, points, degree, source)


def stroud_un_5_3(n):
    degree = 5

    s = sqrt(frac(1, n + 2))
    B = [frac(2 ** (k - n) * (n + 2), n * (k + 1) * (k + 2)) for k in range(1, n + 1)]
    r = [sqrt(frac(k + 2, n + 2)) for k in range(1, n + 1)]
    data = [(B[k], pm(k * [0] + [r[k]] + (n - k - 1) * [s])) for k in range(n)]

    points, weights = untangle(data)
    return UnScheme("Stroud Un 5-3", n, weights, points, degree, source)


def stroud_un_5_4(n):
    degree = 5

    s = sqrt(2 * (n + 2))
    u = sqrt((n + 2 + (n - 1) * s) / n / (n + 2))
    v = sqrt((n + 2 - s) / n / (n + 2))

    data = [(frac(1, 2 ** n * n), fsd(n, (u, 1), (v, n - 1)))]

    points, weights = untangle(data)
    return UnScheme("Stroud Un 5-4", n, weights, points, degree, source)


def stroud_un_7_2(n):
    degree = 7

    A = frac(-(n ** 2), 2 ** (n + 3) * (n + 2))
    B = frac((n + 4) ** 2, 2 ** (n + 3) * n * (n + 2))

    r = sqrt(frac(1, n))
    s = sqrt(frac(5, n + 4))
    t = sqrt(frac(1, n + 4))

    data = [(A, pm(n * [r])), (B, fsd(n, (s, 1), (t, n - 1)))]

    points, weights = untangle(data)
    return UnScheme("Stroud Un 7-2", n, weights, points, degree, source)


__all__ = [
    "stroud_un_3_1",
    "stroud_un_3_2",
    "stroud_un_5_1",
    "stroud_un_5_2",
    "stroud_un_5_3",
    "stroud_un_5_4",
    "stroud_un_7_1",
    "stroud_un_7_2",
    "stroud_un_11_1",
]



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import techreport
from ._helpers import T2Scheme, register

source = techreport(
    authors=["Noel J. Walkington"],
    title="Quadrature on simplices of arbitrary dimension",
    institution="CMU",
    year="2000",
    url="https://www.math.cmu.edu/~nw0z/publications/00-CNA-023/023abs/",
)


def walkington_p5():
    a1, a2 = ((155 + i * sqrt(15)) / 1200 for i in [+1, -1])
    x1, x2 = ((6 + i * sqrt(15)) / 21 for i in [+1, -1])
    d = {"centroid": [[frac(9, 40)]], "d3_aa": [[a1, a2], [x1, x2]]}
    return T2Scheme("Walkington p5", d, 5, source, 2.637e-16)


register([walkington_p5])
